# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import pathlib
import typing

import pydantic

from ..config import DEFAULT_ROBOTO_DIR
from ..domain import datasets
from ..env import resolve_env_variables
from ..http import RobotoClient
from ..logging import default_logger
from ..time import utcnow
from ..updates import MetadataChangeset
from .files import (
    UPLOAD_COMPLETE_FILENAME,
    UPLOAD_IN_PROGRESS_FILENAME,
    UploadAgentConfig,
    UploadCompleteFile,
    UploadConfigFile,
    UploadInProgressFile,
)

logger = default_logger()


DEFAULT_ROBOTO_UPLOAD_FILE = DEFAULT_ROBOTO_DIR / "default_roboto_upload.json"
"""If you put a template JSON file here, it'll be used by `:func:`~create_upload_configs`"""


class UploadAgent:
    __roboto_client: RobotoClient
    __agent_config: UploadAgentConfig

    def __init__(
        self,
        agent_config: UploadAgentConfig,
        roboto_client: typing.Optional[RobotoClient] = None,
    ):
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__agent_config = agent_config

    def create_upload_configs(self):
        directories_to_consider: list[pathlib.Path] = []

        for search_path in self.__agent_config.search_paths:
            if not search_path.is_dir():
                continue

            directories_to_consider.extend(
                [subdir for subdir in search_path.iterdir() if subdir.is_dir()]
            )

        upload_config_file_contents = UploadConfigFile()
        if DEFAULT_ROBOTO_UPLOAD_FILE.is_file():
            try:
                upload_config_file_contents = UploadConfigFile.model_validate_json(
                    resolve_env_variables(DEFAULT_ROBOTO_UPLOAD_FILE.read_text())
                )
                logger.info(
                    "Successfully loaded default config file %s, using it for new uploads.",
                    DEFAULT_ROBOTO_UPLOAD_FILE,
                )
            except Exception:
                logger.warning(
                    "Couldn't parse default config file at %s, using blank contents '{}' instead.",
                    DEFAULT_ROBOTO_UPLOAD_FILE,
                )
        else:
            logger.info(
                "No default upload config file found at %s, using blank contents '{}' instead.",
                DEFAULT_ROBOTO_UPLOAD_FILE,
            )

        for subdir in directories_to_consider:
            if len(list(subdir.iterdir())) == 0:
                logger.info("Directory %s is empty, skipping", subdir)
                continue

            upload_config_file = subdir / self.__agent_config.upload_config_filename
            if upload_config_file.is_file():
                logger.info(
                    "Directory %s already has an upload config file, skipping", subdir
                )
                continue

            upload_in_progress_file = subdir / UPLOAD_IN_PROGRESS_FILENAME
            if upload_in_progress_file.is_file():
                logger.info(
                    "Directory %s has an upload in progress file, skipping", subdir
                )
                continue

            upload_complete_file = subdir / UPLOAD_COMPLETE_FILENAME
            if upload_complete_file.is_file():
                logger.info(
                    "Directory %s has an upload completed file, skipping", subdir
                )
                continue

            upload_config_file.write_text(
                upload_config_file_contents.model_dump_json(indent=2)
            )
            logger.info("Wrote upload config file to %s", upload_config_file)

    def process_uploads(
        self, merge_uploads: bool = False
    ) -> collections.abc.Sequence[datasets.Dataset]:
        """
        If merge is true, everything will be combined into a single dataset. Otherwise, each directory will be uploaded
        to a separate dataset.
        """

        upload_config_files = self.__get_upload_config_files()
        if len(upload_config_files) == 0:
            logger.info(
                "No upload config files found under any search path, nothing to do."
            )
            return []
        else:
            logger.info(
                "Found %d total upload config files across all search paths, starting uploads.",
                len(upload_config_files),
            )

        update_dataset: typing.Optional[datasets.Dataset] = None
        created_datasets: list[datasets.Dataset] = []

        if merge_uploads:
            update_dataset = datasets.Dataset.create(
                roboto_client=self.__roboto_client,
                caller_org_id=self.__agent_config.default_org_id,
            )

            created_datasets.append(update_dataset)

        for upload_config_file, path in upload_config_files:
            uploaded_dataset = self.__handle_upload_config_file(
                file=upload_config_file, path=path, update_dataset=update_dataset
            )
            if uploaded_dataset is not None:
                created_datasets.append(uploaded_dataset)

        return created_datasets

    def __get_upload_config_files(
        self,
    ) -> collections.abc.Collection[tuple[UploadConfigFile, pathlib.Path]]:
        upload_config_files: list[tuple[UploadConfigFile, pathlib.Path]] = []

        for search_path in self.__agent_config.search_paths:
            if not search_path.is_dir():
                logger.error("Search path is not a directory: %s", search_path)
                continue

            found = 0

            logger.info("Scanning '%s' for upload config files", search_path)
            for upload_config_file in search_path.rglob(
                self.__agent_config.upload_config_filename
            ):
                try:
                    parsed_file = UploadConfigFile.model_validate_json(
                        upload_config_file.read_text()
                    )

                    logger.info("Found upload config file: %s", upload_config_file)
                    upload_config_files.append((parsed_file, upload_config_file))
                    found += 1
                except pydantic.ValidationError as exc:
                    logger.error(
                        "Couldn't parse file as valid upload config files: %s",
                        upload_config_file,
                        exc_info=exc,
                    )

            if found == 0:
                logger.info(
                    "No upload config files found for search path: %s", search_path
                )
            else:
                logger.info(
                    "Found %d upload config files under search path: %s",
                    found,
                    search_path,
                )

        return upload_config_files

    def __delete_uploaded_dir_if_safe(self, path: pathlib.Path):
        if not path.is_dir():
            return

        files = [f for f in path.iterdir()]

        if len(files) > 1:
            return

        if len(files) == 1:
            if files[0].name != UPLOAD_COMPLETE_FILENAME:
                return

            files[0].unlink(missing_ok=True)

        path.rmdir()

        logger.info("Cleaned up empty upload directory %s", path)

    def __handle_upload_config_file(
        self,
        file: UploadConfigFile,
        path: pathlib.Path,
        update_dataset: typing.Optional[datasets.Dataset] = None,
    ) -> typing.Optional[datasets.Dataset]:
        """
        If you pass in an update_dataset, it will be used instead of creating a new one, and any
        dataset properties in the upload config file will be applied as an update.

        This will return a dataset if one is created, which means it will NOT return a dataset passed into the
        ``update_dataset`` param.
        """
        dir_to_upload = path.parent

        upload_in_progress_file = dir_to_upload / UPLOAD_IN_PROGRESS_FILENAME
        upload_complete_file = dir_to_upload / UPLOAD_COMPLETE_FILENAME

        if upload_complete_file.is_file():
            try:
                parsed_complete_file = UploadCompleteFile.model_validate_json(
                    upload_complete_file.read_text()
                )
                logger.info(
                    "Found upload-complete file for dataset %s at path %s, skipping upload",
                    parsed_complete_file.dataset_id,
                    upload_complete_file.resolve(),
                )
            except pydantic.ValidationError:
                logger.warning(
                    "Couldn't parse file as a valid upload-complete file. Still ignoring upload to be safe, but "
                    + "it might not have worked: %s",
                    upload_in_progress_file,
                )

            return None

        in_progress_dataset: typing.Optional[datasets.Dataset] = None

        if upload_in_progress_file.is_file():
            try:
                parsed_in_progress_file = UploadInProgressFile.model_validate_json(
                    upload_in_progress_file.read_text()
                )
                in_progress_dataset = datasets.Dataset.from_id(
                    parsed_in_progress_file.dataset_id
                )
                logger.warning(
                    "Found upload-in-progress file for dataset %s at path %s, resuming upload",
                    in_progress_dataset.dataset_id,
                    upload_in_progress_file.resolve(),
                )
            except pydantic.ValidationError:
                logger.warning(
                    "Couldn't parse file as a valid upload-in-progress file, ignoring: %s",
                    upload_in_progress_file,
                )

        dataset: datasets.Dataset
        should_write_in_progress_file: bool

        # Order of figuring out what dataset we're working on:
        # 1. In progress is always the first choice, because if it exists, it means we already made a decision that
        #    we want to stick with, and for some reason the upload failed. Let's not change that.
        # 2. Explicit update
        # 3. Create new
        if in_progress_dataset is not None:
            dataset = in_progress_dataset
            should_write_in_progress_file = False

        elif update_dataset is not None:
            logger.info(
                "Applying update to existing dataset %s for directory: %s",
                update_dataset.dataset_id,
                dir_to_upload,
            )
            dataset = update_dataset
            dataset.update(
                description=file.dataset.description,
                metadata_changeset=MetadataChangeset(
                    put_fields=file.dataset.metadata,
                    put_tags=file.dataset.tags,
                ),
            )
            logger.info("Successfully updated dataset %s", dataset.dataset_id)
            should_write_in_progress_file = True

        else:
            logger.info("Creating a dataset for directory: %s", dir_to_upload)
            dataset = datasets.Dataset.create(
                description=file.dataset.description,
                metadata=file.dataset.metadata,
                tags=file.dataset.tags,
                caller_org_id=file.dataset.org_id,
                roboto_client=self.__roboto_client,
            )
            logger.info("Created dataset %s for path %s", dataset.dataset_id, path)
            should_write_in_progress_file = True

        if should_write_in_progress_file:
            logger.debug(
                "Writing in progress file to %s", upload_in_progress_file.resolve()
            )
            upload_in_progress_file.write_text(
                UploadInProgressFile(
                    version="v1",
                    dataset_id=dataset.dataset_id,
                    started=dataset.record.created,
                ).model_dump_json(indent=2)
            )

        # Default to using the delete-uploaded-files strategy from the agent config file, but override it if it
        # has been explicitly set in the upload file.
        delete_uploaded_files = self.__agent_config.delete_uploaded_files
        if file.upload.delete_uploaded_files is not None:
            delete_uploaded_files = file.upload.delete_uploaded_files

        logger.info(
            "Uploading files from %s to dataset %s. Delete after upload is %s",
            dir_to_upload,
            dataset.dataset_id,
            delete_uploaded_files,
        )

        exclude_patterns = file.upload.exclude_patterns or []
        # Explicitly opt out of uploading the upload-in-progress file.
        exclude_patterns.append(f"**/{UPLOAD_IN_PROGRESS_FILENAME}")

        dataset.upload_directory(
            directory_path=dir_to_upload,
            exclude_patterns=exclude_patterns,
            include_patterns=file.upload.include_patterns,
            delete_after_upload=delete_uploaded_files,
        )

        if path.is_file():
            logger.info("Deleting marker file %s", path)
            path.unlink(missing_ok=True)

        logger.debug(
            "Writing upload complete file to %s", upload_complete_file.resolve()
        )
        upload_complete_file.write_text(
            UploadCompleteFile(
                version="v1", dataset_id=dataset.dataset_id, completed=utcnow()
            ).model_dump_json(indent=2)
        )

        # Delete the in progress file after writing the complete file, because that's when we know for sure that
        # our completion has been signaled to both Roboto service and the local OS.
        upload_in_progress_file.unlink(missing_ok=True)

        # Explicitly write the upload complete file, because it's useful as an "everything is ready" triggers signal,
        # as well as a diagnostic aid.
        dataset.upload_file(upload_complete_file, UPLOAD_COMPLETE_FILENAME)

        logger.info(
            f"Upload completed, view at {self.__roboto_client.frontend_endpoint}/datasets/{dataset.dataset_id}"
        )

        if self.__agent_config.delete_empty_directories:
            self.__delete_uploaded_dir_if_safe(dir_to_upload)

        # If we were passed in an update_dataset, we don't want to return it, because the intention of the return
        # is to be added to an array of created datasets. This one already exists / wasn't created by this call.
        if update_dataset is not None:
            return None

        return dataset



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic

from ...association import Association
from ...sentinels import NotSet, NotSetType
from ...updates import MetadataChangeset


class CreateEventRequest(pydantic.BaseModel):
    """
    Request payload for the Create Event operation.
    """

    associations: list[Association] = pydantic.Field(default_factory=list)
    """
    Datasets, files, topics and message paths which this event relates to. At least one must be provided. All referenced
    datasets, files, topics and message paths must be owned by the same organization.
    """

    description: typing.Optional[str] = None
    """
    An optional human-readable description of the event.
    """

    end_time: int
    """
    The end time of the event, in nanoseconds since epoch (assumed Unix epoch). This can be equal to start_time if
    the event is discrete, but can never be less than start_time.
    """

    metadata: dict[str, typing.Any] = pydantic.Field(
        default_factory=dict,
    )
    """
    Initial key-value pairs to associate with this event for discovery and search.
    """

    name: str
    """
    A brief human-readable name for the event. Many events can have the same name. "Takeoff", "Crash", "CPU Spike",
    "Bad Image Quality", and "Unexpected Left" are a few potential examples.
    """

    start_time: int
    """
    The start time of the event, in nanoseconds since epoch (assumed Unix epoch).
    """

    tags: list[str] = pydantic.Field(default_factory=list)
    """
    Initial tags to associate with this event for discovery and search.
    """


class QueryEventsForAssociationsRequest(pydantic.BaseModel):
    """
    Request payload for the Query Events for Associations operation.
    """

    associations: list[Association]
    """Associations to query events for."""

    page_token: typing.Optional[str] = None
    """Token to use to fetch the next page of results, use None for the first page."""


class UpdateEventRequest(pydantic.BaseModel):
    """
    Request payload for the Update Event operation. Allows any of the mutable fields of an event to be changed.
    """

    description: typing.Union[str, None, NotSetType] = NotSet
    """
    An optional human-readable description of the event.
    """

    metadata_changeset: typing.Union[MetadataChangeset, NotSetType] = NotSet
    """
    Metadata and tag changes to make for this event
    """

    name: typing.Union[str, NotSetType] = NotSet
    """
    A brief human-readable name for the event. Many events can have the same name. "Takeoff", "Crash", "CPU Spike",
    "Bad Image Quality", and "Unexpected Left" are a few potential examples.
    """

    start_time: typing.Union[int, NotSetType] = NotSet
    """
    The start time of the event, in nanoseconds since epoch (assumed Unix epoch).
    """

    end_time: typing.Union[int, NotSetType] = NotSet
    """
    The end time of the event, in nanoseconds since epoch (assumed Unix epoch). This can be equal to start_time if
    the event is discrete, but can never be less than start_time.
    """

    # This is required to get NotSet/NotSetType to serialize appropriately.
    model_config = pydantic.ConfigDict(
        extra="forbid", json_schema_extra=NotSetType.openapi_schema_modifier
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain import actions
from ..command import RobotoCommand
from ..context import CLIContext


def cancel(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    invocation = actions.Invocation.from_id(
        args.invocation_id,
        roboto_client=context.roboto_client,
    )
    invocation.cancel()
    print("Invocation cancelled.")
    return


def cancel_parser(parser: argparse.ArgumentParser):
    parser.add_argument("invocation_id")


cancel_command = RobotoCommand(
    name="cancel",
    logic=cancel,
    setup_parser=cancel_parser,
    command_kwargs={"help": "Cancel invocation."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import json
import typing

from ...domain import actions
from ...query import (
    Comparator,
    Condition,
    ConditionGroup,
    ConditionOperator,
    ConditionType,
    QuerySpecification,
    SortDirection,
)
from ..command import RobotoCommand
from ..common_args import (
    add_action_reference_arg,
    add_org_arg,
)
from ..context import CLIContext


def invocation_serializer(obj: typing.Any) -> typing.Any:
    if isinstance(obj, actions.Invocation):
        return {
            "invocation_id": obj.id,
            "created": obj.created.isoformat(),
            "status": str(obj.current_status),
        }

    # The default behavior for json.dumps(default=) is to throw a TypeError
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


def list_invocations(
    args: argparse.Namespace, context: CLIContext, parser: argparse.ArgumentParser
) -> None:
    conditions = [
        Condition(
            field="provenance.action.name",
            comparator=Comparator.Equals,
            value=args.action.name,
        ),
    ]
    if args.action.digest:
        conditions.append(
            Condition(
                field="provenance.action.digest",
                comparator=Comparator.Equals,
                value=args.action.digest,
            )
        )

    if args.action.owner:
        conditions.append(
            Condition(
                field="provenance.action.org_id",
                comparator=Comparator.Equals,
                value=args.action.owner,
            )
        )

    condition: typing.Optional[ConditionType] = None
    if len(conditions) == 1:
        condition = conditions[0]
    else:
        condition = ConditionGroup(
            operator=ConditionOperator.And,
            conditions=conditions,
        )

    query = QuerySpecification(
        condition=condition,
        sort_by="created",
        sort_direction=SortDirection.Descending,
    )
    matching_invocations = list(
        actions.Invocation.query(
            query,
            owner_org_id=args.org,
            roboto_client=context.roboto_client,
        )
    )
    print(json.dumps(matching_invocations, indent=2, default=invocation_serializer))


def list_invocations_parser(parser: argparse.ArgumentParser):
    add_action_reference_arg(parser)
    add_org_arg(parser=parser)


list_invocations_command = RobotoCommand(
    name="list-invocations",
    logic=list_invocations,
    setup_parser=list_invocations_parser,
    command_kwargs={"help": "List invocations for action."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import base64
import collections.abc
import enum
import http
import http.client
import json
import typing
import urllib.response

import pydantic

from roboto.exceptions import (
    RobotoDomainException,
)

from ..collection_utils import get_by_path
from ..logging import default_logger

logger = default_logger()

Model = typing.TypeVar("Model")
PydanticModel = typing.TypeVar("PydanticModel", bound=pydantic.BaseModel)


class BatchResponseElement(pydantic.BaseModel, typing.Generic[Model]):
    """
    One element of a response to a batch request. This should only ever have data set (in case of a successful
    operation) or error set (in case of a failed operation). For operations that do not return a response, an empty
    (data = None, error = None) Batch Response Element will be effectively equivalent to a single requests 204 No
    Content
    """

    data: typing.Optional[Model] = None
    error: typing.Optional[RobotoDomainException] = None

    @pydantic.field_validator("error", mode="before")
    def validate_error(cls, value: str) -> typing.Optional[RobotoDomainException]:
        try:
            return RobotoDomainException.from_json(json.loads(value))
        except Exception:
            return None

    @pydantic.field_serializer("error")
    def serialize_error(
        self,
        value: typing.Optional[RobotoDomainException],
        info: pydantic.SerializationInfo,
    ) -> typing.Optional[dict[str, typing.Any]]:
        return None if value is None else value.to_dict()


class BatchResponse(pydantic.BaseModel, typing.Generic[Model]):
    """
    The response to a batch request. The responses element contains one response (either success data or failure error)
    per request element, in the order in which the request was sent.
    """

    responses: list[BatchResponseElement[Model]]


class PaginatedList(pydantic.BaseModel, typing.Generic[Model]):
    """
    A list of records pulled from a paginated result set.
    It may be a subset of that result set,
    in which case `next_token` will be set and can be used to fetch the next page.
    """

    items: list[Model]
    # Opaque token that can be used to fetch the next page of results.
    next_token: typing.Optional[str] = None


class StreamedList(pydantic.BaseModel, typing.Generic[Model]):
    """
    A StreamedList differs from a PaginatedList in that it represents a stream of data that is
    in process of being written to. Unlike a result set, which is finite and complete,
    a stream may be infinite, and it is unknown when or if it will complete.
    """

    items: list[Model]
    # Opaque token that can be used to fetch the next page of results.
    last_read: typing.Optional[str]
    # If True, it is known that there are more items to be fetched;
    # use `last_read` as a pagination token to fetch those additional records.
    # If False, it is not known if there are more items to be fetched.
    has_next: bool


class PaginationTokenEncoding(enum.Enum):
    """Pagination token encoding enum"""

    Json = "json"
    Raw = "raw"


class PaginationTokenScheme(enum.Enum):
    """Pagination token scheme enum"""

    V1 = "v1"


class PaginationToken:
    """
    A pagination token that can be treated as a truly opaque token by clients,
    with support for evolving the token format over time.
    """

    __scheme: PaginationTokenScheme
    __encoding: PaginationTokenEncoding
    __data: typing.Any

    @staticmethod
    def empty() -> "PaginationToken":
        return PaginationToken(
            PaginationTokenScheme.V1, PaginationTokenEncoding.Raw, None
        )

    @staticmethod
    def encode(data: str) -> str:
        """Base64 encode the data and strip all trailing padding ("=")."""
        return (
            base64.urlsafe_b64encode(data.encode("utf-8")).decode("utf-8").rstrip("=")
        )

    @staticmethod
    def decode(data: str) -> str:
        """Base64 decode the data, adding back any trailing padding ("=") as necessary to make data properly Base64."""
        while len(data) % 4 != 0:
            data += "="
        return base64.urlsafe_b64decode(data).decode("utf-8")

    @classmethod
    def from_token(cls, token: typing.Optional[str]) -> "PaginationToken":
        if token is None:
            return PaginationToken.empty()
        try:
            decoded = PaginationToken.decode(token)
            if not decoded.startswith(PaginationTokenScheme.V1.value):
                logger.error("Invalid pagination token scheme %s", decoded)
                raise ValueError("Invalid pagination token scheme")
            scheme, encoding, data = decoded.split(":", maxsplit=2)
            pagination_token_scheme = PaginationTokenScheme(scheme)
            pagination_token_encoding = PaginationTokenEncoding(encoding)
            return cls(
                pagination_token_scheme,
                pagination_token_encoding,
                (
                    json.loads(data)
                    if pagination_token_encoding == PaginationTokenEncoding.Json
                    else data
                ),
            )
        except Exception as e:
            logger.error(f"Invalid pagination token {token}", exc_info=e)
            raise ValueError("Invalid pagination token format") from None

    def __init__(
        self,
        scheme: PaginationTokenScheme,
        encoding: PaginationTokenEncoding,
        data: typing.Any,
    ):
        self.__scheme = scheme
        self.__encoding = encoding
        self.__data = data

    def __len__(self):
        return len(str(self)) if self.__data else 0

    def __str__(self):
        return self.to_token()

    @property
    def data(self) -> typing.Any:
        return self.__data

    def to_token(self) -> str:
        data = (
            json.dumps(self.__data)
            if self.__encoding == PaginationTokenEncoding.Json
            else self.__data
        )
        return PaginationToken.encode(
            f"{self.__scheme.value}:{self.__encoding.value}:{data}"
        )


class HttpResponse:
    __response: urllib.response.addinfourl

    def __init__(self, response: urllib.response.addinfourl) -> None:
        super().__init__()
        self.__response = response

    @property
    def readable_response(self) -> urllib.response.addinfourl:
        return self.__response

    @property
    def status(self) -> http.HTTPStatus:
        status_code = self.__response.status
        if status_code is None:
            raise RuntimeError("Response has no status code")
        return http.HTTPStatus(int(status_code))

    @property
    def headers(self) -> typing.Optional[dict[str, str]]:
        return dict(self.__response.headers.items())

    def to_paginated_list(
        self, record_type: typing.Type[PydanticModel]
    ) -> PaginatedList[PydanticModel]:
        unmarshalled = self.to_dict(json_path=["data"])
        return PaginatedList(
            items=[record_type.model_validate(item) for item in unmarshalled["items"]],
            next_token=unmarshalled["next_token"],
        )

    def to_record(self, record_type: typing.Type[PydanticModel]) -> PydanticModel:
        return record_type.model_validate(self.to_dict(json_path=["data"]))

    def to_record_list(
        self, record_type: typing.Type[PydanticModel]
    ) -> collections.abc.Sequence[PydanticModel]:
        return [
            record_type.model_validate(item)
            for item in self.to_dict(json_path=["data"])
        ]

    def to_string_list(self) -> list[str]:
        return [str(item) for item in self.to_dict(json_path=["data"])]

    def to_dict(self, json_path: typing.Optional[list[str]] = None) -> typing.Any:
        with self.__response:
            unmarsalled = json.loads(self.__response.read().decode("utf-8"))
            if json_path is None:
                return unmarsalled

            return get_by_path(unmarsalled, json_path)

    def to_string(self):
        with self.__response:
            return self.__response.read().decode("utf-8")



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


"""
SerDe: Serialization + Deserialization

This module contains utility functions for common serialization and deserialization scenarios used
throughout the Roboto python codebase.
"""

from .arrays import safe_access_array
from .dicts import (
    case_insensitive_get,
    pydantic_jsonable_dict,
    pydantic_jsonable_dict_map,
    pydantic_jsonable_dicts,
    safe_dict_drill,
)
from .paths import (
    exclude_patterns_to_spec,
    git_paths_match,
    git_paths_to_spec,
)

__all__ = [
    "pydantic_jsonable_dict",
    "pydantic_jsonable_dicts",
    "pydantic_jsonable_dict_map",
    "safe_dict_drill",
    "case_insensitive_get",
    "exclude_patterns_to_spec",
    "git_paths_to_spec",
    "git_paths_match",
    "safe_access_array",
]



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
import collections.abc
import enum
import json
import shlex
import typing

from ...domain import actions
from ...sentinels import NotSetType, null
from ..command import KeyValuePairsAction
from ..validation import (
    pydantic_validation_handler,
)


class ActionReferenceParser(argparse.Action):
    """Parse an action reference string into its org (if given) and action name components."""

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: typing.Union[str, collections.abc.Sequence[str], None],
        option_string: typing.Optional[str] = None,
    ):
        if not isinstance(values, str):
            raise ValueError("ActionReference must be a string")

        org_name_parts = values.split("/", maxsplit=1)
        name_part = org_name_parts[-1]
        name_digest_parts = name_part.split("@", maxsplit=1)

        parse_result = dict()

        if len(org_name_parts) == 2:
            # Org is explicit: Action belongs to the org specified in the reference
            # Example: "some-org/an-action"
            action_owner = org_name_parts[0]
            parse_result["owner"] = action_owner

        if len(name_digest_parts) == 2:
            action_name, action_digest = name_digest_parts
            parse_result["name"] = action_name
            parse_result["digest"] = action_digest
        elif len(name_digest_parts) == 1:
            action_name = name_digest_parts[0]
            parse_result["name"] = action_name

        setattr(
            namespace, self.dest, actions.ActionReference.model_validate(parse_result)
        )


def add_action_reference_arg(
    parser: argparse.ArgumentParser,
    positional: bool = True,
    required: bool = True,
    arg_help: str = "Partially or fully qualified reference to Action on which to operate.",
    arg_name="action",
) -> None:
    arg = arg_name if positional else f"--{arg_name}"
    kwargs: dict[str, typing.Any] = {
        "metavar": "<action_name> | <action_name>@<digest> | <org>/<action_name> | <org>/<action_name>@<digest>",  # noqa: E501
        "action": ActionReferenceParser,
        "help": arg_help,
    }
    if not positional and required:
        kwargs["required"] = True

    parser.add_argument(
        arg,
        **kwargs,
    )


class ActionParameterArg(argparse.Action):
    """Parse an action parameter string an ActionParameter value object."""

    METAVAR: typing.ClassVar[str] = (
        "'name=<name>|required=<true|false>|default=<value>|description=<description>'"  # noqa: E501
    )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: typing.Union[str, collections.abc.Sequence[str], None],
        option_string: typing.Optional[str] = None,
    ):
        if not values:
            return

        parameters = []
        for parameter_str in values:
            parts = parameter_str.split("|")
            parse_result = {}
            for part in parts:
                trimmed = part.strip()
                if not trimmed:
                    continue

                parts = trimmed.split("=", maxsplit=1)
                if not len(parts) == 2:
                    raise argparse.ArgumentError(
                        self,
                        f"Unable to parse action parameter '{parameter_str}'. "
                        f"Parameter '{part}' is not parsable as a <key>=<value> pair.",
                    )

                key = parts[0].strip()
                value = parts[1].strip()
                parsed_value: typing.Any = value
                if key == "required":
                    parsed_value = parsed_value.lower() == "true"
                else:
                    try:
                        parsed_value = json.loads(parsed_value)
                    except Exception:
                        pass
                parse_result[key] = parsed_value

            with pydantic_validation_handler("Action parameter"):
                action_parameter = actions.ActionParameter.model_validate(parse_result)

            parameters.append(action_parameter)

        existing_parameters = getattr(namespace, self.dest, None)
        if not existing_parameters:
            existing_parameters = []
        setattr(namespace, self.dest, [*existing_parameters, *parameters])


class ActionTimeoutArg(argparse.Action):
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: typing.Union[str, collections.abc.Sequence[str], NotSetType, None],
        option_string: typing.Optional[str] = None,
    ):
        if values is None or isinstance(values, NotSetType):
            setattr(namespace, self.dest, values)
            return

        if not isinstance(values, str):
            raise argparse.ArgumentError(self, "Can only specify one timeout value.")

        timeout: typing.Optional[int] = None
        try:
            timeout = int(values)
        except ValueError:
            raise argparse.ArgumentError(
                self,
                f"Invalid value for timeout: '{values}'. Timeout must be parseable as an integer.",
            )
        if timeout <= 0:
            raise argparse.ArgumentError(
                self,
                f"Invalid value for timeout: {timeout}. Timeout must be greater than 0.",
            )

        setattr(namespace, self.dest, timeout)


class DockerInstructionForm(enum.Enum):
    """The form of a CMD instruction."""

    Exec = "exec"
    Shell = "shell"


def parse_compute_requirements(
    args: argparse.Namespace,
    defaults: typing.Optional[actions.ComputeRequirements] = None,
) -> typing.Optional[actions.ComputeRequirements]:
    if args.vcpu is None and args.memory is None and args.storage is None:
        return None

    defaults = defaults if defaults else actions.ComputeRequirements()

    with pydantic_validation_handler("compute requirements"):
        kwargs = {
            key: value
            for key, value in [
                ("vCPU", args.vcpu if args.vcpu else defaults.vCPU),
                ("memory", args.memory if args.memory else defaults.memory),
                ("storage", args.storage if args.storage else defaults.storage),
            ]
            if value is not None
        }
        if not kwargs:
            return None
        return actions.ComputeRequirements.model_validate(kwargs)


def add_compute_requirements_args(parser: argparse.ArgumentParser) -> None:
    resource_requirements_group = parser.add_argument_group(
        "Resource requirements",
        "Specify required compute resources.",
    )
    resource_requirements_group.add_argument(
        "--vcpu",
        required=False,
        type=int,
        choices=[256, 512, 1024, 2048, 4096, 8192, 16384],
        help="CPU units to dedicate to action invocation. Defaults to 512 (0.5vCPU).",
    )

    resource_requirements_group.add_argument(
        "--memory",
        required=False,
        type=int,
        help=(
            "Memory (in MiB) to dedicate to action invocation. Defaults to 1024 (1 GiB). "
            "Supported values range from 512 (0.5 GiB) to 122880 (120 GiB). "
            "Supported values are tied to selected vCPU resources. See documentation for more information."
        ),
    )

    resource_requirements_group.add_argument(
        "--storage",
        required=False,
        type=int,
        help=(
            "Ephemeral storage (in GiB) to dedicate to action invocation. Defaults to 21 GiB. "
            "Supported values range from 21 to 200, inclusive."
        ),
    )

    # Placeholder
    resource_requirements_group.add_argument(
        "--gpu",
        required=False,
        default=False,
        action="store_true",
        help=(
            "This is a placeholder; it currently does nothing. "
            "In the future, setting this option will invoke the action in a GPU-enabled compute environment."
        ),
    )


def parse_container_overrides(
    args: argparse.Namespace,
    defaults: typing.Optional[actions.ContainerParameters] = None,
) -> typing.Optional[actions.ContainerParameters]:
    if (
        args.entry_point is None
        and args.command is None
        and args.workdir is None
        and args.env is None
    ):
        return None

    defaults = defaults if defaults else actions.ContainerParameters()

    with pydantic_validation_handler("container parameters"):
        entry_point: typing.Union[list[str], object] = defaults.entry_point
        if args.entry_point is null:
            entry_point = null
        elif args.entry_point is not None:
            entry_point = [args.entry_point]

        command: typing.Union[list[str], object] = defaults.command
        if args.command is null:
            command = null
        elif args.command is not None:
            command_form = DockerInstructionForm(args.command_form)
            command = []
            if command_form == DockerInstructionForm.Exec and len(args.command):
                lexxer = shlex.shlex(args.command, posix=True, punctuation_chars=True)
                lexxer.whitespace_split = True
                command = list(lexxer)
            else:
                command = [args.command]

        kwargs = {
            key: value
            for key, value in [
                ("entry_point", entry_point),
                ("command", command),
                ("workdir", args.workdir if args.workdir else defaults.workdir),
                ("env_vars", args.env if args.env else defaults.env_vars),
            ]
            if value is not None
        }
        if not kwargs:
            return None
        return actions.ContainerParameters.model_validate(
            {key: value if value is not null else None for key, value in kwargs.items()}
        )


def add_container_parameters_args(parser: argparse.ArgumentParser) -> None:
    group = parser.add_argument_group(
        "Container parameters",
        "Specify parameters to pass to the action's Docker container at runtime.",
    )

    group.add_argument(
        "--entrypoint",
        required=False,
        type=lambda s: s if s != "null" else null,
        dest="entry_point",
        help=(
            "Container ENTRYPOINT override."
            ' Supports passing empty string ("") as an override, which unsets the ENTRYPOINT specified in the docker image.'  # noqa: E501
            " If updating or invoking action which has existing ENTRYPOINT override, specify ``null`` to remove the override."  # noqa: E501
            " Refer to docker documentation for more: "
            "https://docs.docker.com/engine/reference/builder/#entrypoint"
            " and https://docs.docker.com/engine/reference/run/#entrypoint-default-command-to-execute-at-runtime"
        ),
    )

    group.add_argument(
        "--command",
        required=False,
        type=lambda s: s if s != "null" else null,
        dest="command",
        help=(
            "Container CMD override."
            " If updating or invoking action which has existing CMD override, specify ``null`` to remove the override."
            " Refer to docker documentation for more: "
            "https://docs.docker.com/engine/reference/builder/#cmd and"
            " https://docs.docker.com/engine/reference/run/#cmd-default-command-or-options"
        ),
    )

    group.add_argument(
        "--command-form",
        required=False,
        choices=[form.value for form in DockerInstructionForm],
        default=DockerInstructionForm.Exec.value,
        dest="command_form",
        help=(
            "In ``exec`` form, the provided ``--command`` str is split into a list of strings"
            ' (e.g., ``--command "-c \'print(123)\'"`` is parsed as ``["-c", "print(123)"]``).'
            " In ``shell`` form, the provided ``--command`` str is not split"
            " (e.g., ``--command \"python -c 'print(123)'\"`` is parsed as ``[\"python -c 'print(123)'\"]``)."
        ),
    )

    group.add_argument(
        "--workdir",
        required=False,
        type=lambda s: s if s != "null" else null,
        dest="workdir",
        help=(
            "If updating, specify ``null`` to clear existing workdir."
            " Refer to docker documentation for more: https://docs.docker.com/engine/reference/run/#workdir"
        ),
    )

    group.add_argument(
        "--env",
        required=False,
        metavar="KEY=VALUE",
        nargs="*",
        action=KeyValuePairsAction,
        help=(
            "Zero or more ``<key>=<value>`` formatted pairs to set as container ENV vars. "
            "Do not use ENV vars for secrets (such as API keys). "
            "See documentation: https://docs.docker.com/engine/reference/run/#env-environment-variables"
        ),
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import collections.abc
import typing
import urllib.parse

from ...http import RobotoClient
from .operations import (
    CreateCommentRequest,
    UpdateCommentRequest,
)
from .record import (
    CommentEntityType,
    CommentRecord,
)


class Comment:
    """
    Comments can be made on a number of Roboto resources, like datasets, files, action invocations, and more.
    They support ``@<user_id>`` mention syntax that you can use to notify others of the comment.
    """

    __record: CommentRecord
    __roboto_client: RobotoClient

    @classmethod
    def create(
        cls,
        comment_text: str,
        entity_id: str,
        entity_type: CommentEntityType,
        roboto_client: typing.Optional[RobotoClient] = None,
        caller_org_id: typing.Optional[str] = None,
    ) -> "Comment":
        roboto_client = RobotoClient.defaulted(roboto_client)

        request = CreateCommentRequest(
            entity_type=entity_type,
            entity_id=entity_id,
            comment_text=comment_text,
        )

        record = roboto_client.post(
            "v1/comments", data=request, caller_org_id=caller_org_id
        ).to_record(CommentRecord)

        return cls(record, roboto_client)

    @classmethod
    def from_id(
        cls,
        comment_id: str,
        owner_org_id: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> "Comment":
        roboto_client = RobotoClient.defaulted(roboto_client)
        record = roboto_client.get(
            f"v1/comments/{comment_id}", owner_org_id=owner_org_id
        ).to_record(CommentRecord)
        return cls(record, roboto_client)

    @classmethod
    def for_entity(
        cls,
        entity_type: CommentEntityType,
        entity_id: str,
        owner_org_id: typing.Optional[str] = None,
        page_token: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> tuple[collections.abc.Sequence["Comment"], typing.Optional[str]]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        query_params: dict[str, typing.Any] = {}
        if page_token:
            query_params["page_token"] = page_token

        response_page = roboto_client.get(
            f"v1/comments/{entity_type}/{entity_id}",
            query=query_params,
            owner_org_id=owner_org_id,
        ).to_paginated_list(CommentRecord)

        return [
            cls(record, roboto_client) for record in response_page.items
        ], response_page.next_token

    @classmethod
    def for_entity_type(
        cls,
        entity_type: CommentEntityType,
        owner_org_id: typing.Optional[str] = None,
        page_token: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> tuple[collections.abc.Sequence["Comment"], typing.Optional[str]]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        query_params: dict[str, typing.Any] = {}
        if page_token:
            query_params["page_token"] = page_token

        response_page = roboto_client.get(
            f"v1/comments/type/{entity_type}",
            query=query_params,
            owner_org_id=owner_org_id,
        ).to_paginated_list(CommentRecord)

        return [
            cls(record, roboto_client) for record in response_page.items
        ], response_page.next_token

    @classmethod
    def for_user(
        cls,
        user_id: str,
        owner_org_id: typing.Optional[str] = None,
        page_token: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> tuple[collections.abc.Sequence["Comment"], typing.Optional[str]]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        query_params: dict[str, typing.Any] = {}
        if page_token:
            query_params["page_token"] = page_token

        quoted_user_id = urllib.parse.quote(user_id, safe="")
        response_page = roboto_client.get(
            f"v1/comments/user/{quoted_user_id}",
            query=query_params,
            owner_org_id=owner_org_id,
        ).to_paginated_list(CommentRecord)

        return [
            cls(record, roboto_client) for record in response_page.items
        ], response_page.next_token

    @classmethod
    def recent_for_org(
        cls,
        owner_org_id: typing.Optional[str] = None,
        page_token: typing.Optional[str] = None,
        roboto_client: typing.Optional[RobotoClient] = None,
    ) -> tuple[collections.abc.Sequence["Comment"], typing.Optional[str]]:
        roboto_client = RobotoClient.defaulted(roboto_client)

        query_params: dict[str, typing.Any] = {}
        if page_token:
            query_params["page_token"] = page_token

        response_page = roboto_client.get(
            "v1/comments/recent", query=query_params, owner_org_id=owner_org_id
        ).to_paginated_list(CommentRecord)

        return [
            cls(record, roboto_client) for record in response_page.items
        ], response_page.next_token

    def __init__(
        self, record: CommentRecord, roboto_client: typing.Optional[RobotoClient] = None
    ) -> None:
        self.__roboto_client = RobotoClient.defaulted(roboto_client)
        self.__record = record

    def __repr__(self) -> str:
        return self.__record.model_dump_json()

    @property
    def comment_id(self) -> str:
        return self.__record.comment_id

    @property
    def created_by(self) -> str:
        return self.__record.created_by

    @property
    def record(self) -> CommentRecord:
        return self.__record

    def delete_comment(self) -> None:
        self.__roboto_client.delete(f"/v1/comments/{self.comment_id}")

    def update_comment(self, comment_text: str) -> "Comment":
        self.__record = self.__roboto_client.put(
            f"/v1/comments/{self.comment_id}",
            data=UpdateCommentRequest(
                comment_text=comment_text,
            ),
        ).to_record(CommentRecord)

        return self



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.collections import Collection
from ..command import RobotoCommand
from ..context import CLIContext
from .shared_helpdoc import COLLECTION_ID_HELP


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    collection = Collection.from_id(
        collection_id=args.collection_id,
        roboto_client=context.roboto_client,
    )
    for change in collection.changes(
        from_version=args.from_version, to_version=args.to_version
    ):
        print(change.model_dump_json(indent=2))


def show_setup_parser(parser):
    parser.add_argument("collection_id", type=str, help=COLLECTION_ID_HELP)
    parser.add_argument(
        "--from-version",
        type=int,
        required=False,
        help="A collection version to use as the starting point in change-set listing. Defaults to initial.",
    )
    parser.add_argument(
        "--to-version",
        type=int,
        required=False,
        help="A collection version to use as the end point in change-set listing. Defaults to latest.",
    )


changes_command = RobotoCommand(
    name="changelog",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={
        "help": "Provides a changelog for the revisions of a given collection."
    },
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import enum


class RobotoRegion(str, enum.Enum):
    """
    The geographic region of a Roboto resource. Used when configuring org-level default behavior for data storage, in
    order to ensure that data is close to your users.
    """

    US_WEST = "us-west"
    US_GOV_WEST = "us-gov-west"
    US_EAST = "us-east"
    EU_CENTRAL = "eu-central"



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse
from typing import Any, Callable, Optional

from ...exceptions import (
    RobotoDomainException,
    RobotoNoOrgProvidedException,
)
from ..argparse import SortingHelpFormatter
from ..context import CLIContext
from ..terminal import print_error_and_exit


class RobotoCommand(object):
    _name: str
    _logic: Callable[
        [Any, CLIContext, argparse.ArgumentParser], None
    ]  # Args, CLIContext
    _inner_setup_parser: Optional[Callable[[Any], None]]  # Parser
    _command_kwargs: Optional[dict[str, Any]]

    def __init__(
        self,
        name: str,
        logic: Callable[[Any, CLIContext, argparse.ArgumentParser], None],
        command_kwargs: Optional[dict],
        setup_parser: Optional[Callable[[Any], None]] = None,
    ):
        self._name = name
        self._logic = logic
        self._inner_setup_parser = setup_parser
        self._command_kwargs = command_kwargs

    @property
    def name(self):
        return self._name

    @property
    def command_kwargs(self):
        if self._command_kwargs is not None:
            return self._command_kwargs
        return {}

    def setup_parser(
        self, parser: argparse.ArgumentParser, context: Optional[CLIContext]
    ):
        def context_aware_logic(args):
            if not context:
                raise ValueError(
                    "CLIContext is not set. "
                    "This is a programming error, and should be fixed by the Roboto development team."
                )
            try:
                return self._logic(args, context, parser)
            except RobotoNoOrgProvidedException:
                parser.error(
                    "User with 0 or 2+ orgs must explicitly provide --org to org-bound operations. "
                    + "This can also be achieved by setting a 'ROBOTO_ORG_ID' environment variable, "
                    + "whose value will be used by default."
                )
            except RobotoDomainException as exc:
                print_error_and_exit(f"{exc.__class__.__name__}: {exc}")

        parser.set_defaults(func=context_aware_logic)
        if self._inner_setup_parser is not None:
            self._inner_setup_parser(parser)


class RobotoCommandSet(object):
    commands: list[RobotoCommand]
    name: str

    __help: str
    __context: Optional[CLIContext]

    def __init__(self, name, help, commands):
        self.name = name
        self.__help = help
        self.commands = commands

    def add_to_subparsers(
        self,
        parent_subparsers: Any,
        context: Optional[CLIContext],
    ) -> None:
        self.__context = context

        command_set_parser = parent_subparsers.add_parser(self.name, help=self.__help)
        subparsers = command_set_parser.add_subparsers()
        subparsers.required = True

        for command in self.commands:
            command_parser = subparsers.add_parser(
                command.name,
                formatter_class=SortingHelpFormatter,
                **command.command_kwargs,
            )
            command.setup_parser(command_parser, context)

    def sort_commands(self):
        self.commands.sort(key=lambda command: command.name)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import typing

import pydantic

from ...regionalization import RobotoRegion
from .org_records import OrgRoleName, OrgStatus


class CreateOrgRequest(pydantic.BaseModel):
    """
    Request payload to create an organization
    """

    name: str
    bind_email_domain: bool = False
    data_region: RobotoRegion = RobotoRegion.US_WEST


class OrgRecordUpdates(pydantic.BaseModel):
    """
    Payload to update an organization
    """

    name: typing.Optional[str] = None
    status: typing.Optional[OrgStatus] = None


class UpdateOrgRequest(pydantic.BaseModel):
    """
    Request payload to update an organization
    """

    updates: OrgRecordUpdates


class UpdateOrgUserRequest(pydantic.BaseModel):
    """
    Request payload to update an organization user
    """

    add_roles: typing.Optional[list[OrgRoleName]] = None
    remove_roles: typing.Optional[list[OrgRoleName]] = None

    @pydantic.model_validator(mode="after")
    def check_some_updates_present(self) -> "UpdateOrgUserRequest":
        if len(self.add_roles or []) == 0 and len(self.remove_roles or []) == 0:
            raise ValueError(
                "At least one role should be included in add_roles or remove_roles"
            )

        return self


class RemoveUserFromOrgRequest(pydantic.BaseModel):
    """
    Request payload to remove a user from an organization
    """

    user_id: str


# Deprecated
class ModifyRoleForUserRequest(pydantic.BaseModel):
    """
    Request payload to modify the role for a user in an organization
    """

    user_id: str
    role_name: OrgRoleName


class BindEmailDomainRequest(pydantic.BaseModel):
    """
    Request payload to bind an email domain to an organization
    """

    email_domain: str


class InviteUserRequest(pydantic.BaseModel):
    """
    Request payload to invite a user to an organization
    """

    invited_user_id: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from typing import Optional

import pydantic

from .record import (
    NotificationChannel,
    ReadStatus,
)
from .validator import LifecycleStatusValidator


class UpdateNotificationRequest(pydantic.BaseModel, LifecycleStatusValidator):
    """Request payload to update a notification"""

    notification_id: str
    read_status: Optional[ReadStatus] = None
    lifecycle_status: Optional[dict[NotificationChannel, str]] = None



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import enum
import typing
import urllib.parse

import pydantic

# Python 3.8/3.9 compatible import of TypeAlias
try:
    from typing import TypeAlias
except ImportError:
    try:
        from typing_extensions import TypeAlias
    except ImportError:
        pass


class FileStatus(str, enum.Enum):
    """
    At the initiation of an upload, a file's status is marked as ``Reserved``.
    Once the upload completes successfully, it is set to ``Available``,
    and is made visible when listing files in a dataset or searching files.
    A file's status may be temporarily set to ``Deleted`` if its deletion is in progress.
    """

    Available = "available"
    Deleted = "deleted"
    Reserved = "reserved"


class IngestionStatus(str, enum.Enum):
    """
    A file is considered ``Ingested`` if it has any associated topic records.
    File ingestion happens as a post-upload processing step, and Roboto supports
    a number of common robotics log formats (such as ROS bags, MCAP files, and ULOG files) out-of-the-box.

    An ingested file generally has first-class visualization support.
    """

    NotIngested = "not_ingested"
    Ingested = "ingested"


class FileStorageType(str, enum.Enum):
    """
    File storage type enum
    """

    S3Imported = "imported"
    """
    This file was imported from a read-only customer bucket.
    """

    S3Uploaded = "uploaded"
    """
    This file was explicitly uploaded to either a Roboto managed bucket or a read/write customer bring-your-own bucket.
    """


class FileRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a file.
    """

    association_id: (
        str  # e.g. dataset_id, collection_id, etc.; GSI PK of "association_id" index.
    )
    created: datetime.datetime
    created_by: str = ""
    description: typing.Optional[str] = None
    device_id: typing.Optional[str] = None
    file_id: str
    ingestion_status: IngestionStatus = IngestionStatus.NotIngested
    metadata: dict[str, typing.Any] = pydantic.Field(default_factory=dict)
    modified: datetime.datetime
    modified_by: str
    org_id: str
    origination: str = ""  # Defaulted for compatibility
    relative_path: (
        str  # path relative to some common prefix. Used as local path when downloaded.
    )
    size: int  # bytes
    status: FileStatus = FileStatus.Available
    storage_type: FileStorageType = FileStorageType.S3Uploaded
    tags: list[str] = pydantic.Field(default_factory=list)
    upload_id: str = "NO_ID"  # Defaulted for backwards compatability
    uri: str
    version: int

    @property
    def bucket(self) -> str:
        parsed_uri = urllib.parse.urlparse(self.uri)
        return parsed_uri.netloc

    @property
    def key(self) -> str:
        parsed_uri = urllib.parse.urlparse(self.uri)
        return parsed_uri.path.lstrip("/")


class FileTag(enum.Enum):
    """
    FileTag enum
    """

    DatasetId = "dataset_id"
    OrgId = "org_id"
    # Path to file relative to common prefix
    CommonPrefix = "common_prefix"
    TransactionId = "transaction_id"


class S3Credentials(typing.TypedDict):
    """
    This interface is driven by botocore.credentials.RefreshableCredentials
    """

    access_key: str
    secret_key: str
    token: str
    region: str
    expiry_time: typing.Optional[str]


CredentialProvider: TypeAlias = typing.Callable[[], S3Credentials]


class DirectoryRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a dataset directory.
    """

    name: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import datetime
from enum import Enum

import pydantic


class CommentEntityType(str, Enum):
    """
    Roboto entities that are comment-able.
    """

    Action = "action"
    Collection = "collection"
    Dataset = "dataset"
    File = "file"
    Invocation = "invocation"
    Trigger = "trigger"


class CommentRecord(pydantic.BaseModel):
    """
    A wire-transmissible representation of a comment.
    """

    # Primary key, defined in CDK
    org_id: str  # partition key
    comment_id: str
    entity_type: CommentEntityType
    entity_id: str
    # Persisted as ISO 8601 string in UTC
    created: datetime.datetime
    created_by: str
    comment_text: str
    mentions: list[str] = pydantic.Field(default_factory=list)
    # Persisted as ISO 8601 string in UTC
    modified: datetime.datetime
    modified_by: str

    # for v2
    # parent_comment_id: str



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.


from typing import Optional

import pathspec


def git_pathspec_escape_path(path: str) -> str:
    """
    Turns a literal path into a gitignore glob that matches the literal path.
    """
    escaped = path.replace("[", "\\[")
    escaped = escaped.replace("]", "\\]")
    return escaped


def git_paths_to_spec(paths: list[str]) -> pathspec.PathSpec:
    return pathspec.PathSpec.from_lines(pathspec.patterns.GitWildMatchPattern, paths)


def git_paths_match(
    include_patterns: Optional[list[str]],
    exclude_patterns: Optional[list[str]],
    file: str,
) -> bool:
    # Include patterns are provided, and the file isn't included, ignore it
    if include_patterns is not None:
        if not git_paths_to_spec(include_patterns).match_file(file):
            return False

    # Exclude pattern is provided, and the file is included, ignore it
    if exclude_patterns is not None:
        if git_paths_to_spec(exclude_patterns).match_file(file):
            return False

    return True


def exclude_patterns_to_spec(
    exclude_patterns: Optional[list[str]] = None,
) -> Optional[pathspec.PathSpec]:
    return (
        None
        if exclude_patterns is None
        else pathspec.GitIgnoreSpec.from_lines(exclude_patterns)
    )



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import decimal
import typing

from roboto.types import UserMetadata


def field_serializer_user_metadata(value: dict[str, typing.Any]) -> UserMetadata:
    for k, v in value.items():
        if type(v) in [bool, int, float, str]:
            continue
        elif type(v) is decimal.Decimal:
            value[k] = float(v)
        else:
            raise ValueError(
                f"Illegal metadata element with key '{k}',  type {type(v)}"
            )

    return value



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.collections import Collection
from ..command import RobotoCommand
from ..context import CLIContext
from .shared_helpdoc import COLLECTION_ID_HELP


def delete(args, context: CLIContext, parser: argparse.ArgumentParser):
    Collection.from_id(
        collection_id=args.collection_id,
        roboto_client=context.roboto_client,
    ).delete()
    print(f"Deleted collection {args.collection_id}")


def delete_setup_parser(parser):
    parser.add_argument("collection_id", type=str, help=COLLECTION_ID_HELP)


delete_command = RobotoCommand(
    name="delete",
    logic=delete,
    setup_parser=delete_setup_parser,
    command_kwargs={"help": "Delete a collection."},
)



# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import argparse

from ...domain.collections import (
    Collection,
    CollectionContentMode,
)
from ..command import RobotoCommand
from ..context import CLIContext
from .shared_helpdoc import (
    COLLECTION_ID_HELP,
    COLLECTION_VERSION_HELP,
    CONTENT_MODE_HELP,
)


def show(args, context: CLIContext, parser: argparse.ArgumentParser):
    content_mode = [x for x in CollectionContentMode if x.value == args.content_mode][0]

    collection = Collection.from_id(
        collection_id=args.collection_id,
        version=args.collection_version,
        roboto_client=context.roboto_client,
        content_mode=content_mode,
    )
    print(collection.record.model_dump_json(indent=2))


def show_setup_parser(parser):
    parser.add_argument("collection_id", type=str, help=COLLECTION_ID_HELP)
    parser.add_argument(
        "-v",
        "--collection-version",
        type=int,
        required=False,
        help=COLLECTION_VERSION_HELP,
    )
    parser.add_argument(
        "--content-mode",
        type=str,
        choices=[mode.value for mode in CollectionContentMode],
        help=CONTENT_MODE_HELP,
        default=CollectionContentMode.References.value,
    )


show_command = RobotoCommand(
    name="show",
    logic=show,
    setup_parser=show_setup_parser,
    command_kwargs={"help": "Show information about a specific collection."},
)
